#include <Wire.h> //protocolo I2C
#include "BH1750.h" //libreria especifica del sensor de luz
#include <OneWire.h> // 1-Wire library
#include <DallasTemperature.h> // DallasTemperature library
#include <DHT.h> //Libreria del sensor de humedad

#define DEBUG

#define ONE_WIRE_BUS 2 // Los datos del 1 wire se reciben por el pin 2 del Arduino
#define PIN_PH A0 //Pin analógico para leer los valores de pH
#define DHTPIN 8 // Definimos el pin digital donde se conecta el sensor de humedad
#define DHTTYPE DHT11 // Definimos el tipo de sensor
#define SISTEMALUZ 9 // definir el pin al que irá conectado el sistema de iluminación

BH1750 lightMeter;

OneWire oneWire(ONE_WIRE_BUS); // Perparar una instancia del 1-Wire para comunicarse con cualquier dispositivo 1-Wire, en nuestro caso el Dallas Temperature
DHT dht(DHTPIN, DHTTYPE);  // Inicializamos el sensor DHT11

DallasTemperature sensors(&oneWire); // Pasar la referencia del 1-Wire al Dallas Temperature
DeviceAddress tempSensorAgua = {0x28, 0xFF, 0x96, 0xF0, 0x00, 0x17, 0x03, 0x72};
DeviceAddress tempSensorAire = {0x28, 0xFF, 0xCD, 0xBD, 0x00, 0x17, 0x03, 0xEA};

bool luxometreOK, // Estado del sensor BH1750
     agua;        // Estado del nivel del agua

int  luz;         // Luminosidad leida por el sensor de luz

float temperaturaAgua = 0,   // Temperatura del agua
      temperaturaAire = 0,   // Temperatura del aire
      pH = 0,                // Nivel de pH
      humedad = 0,           // Humedad relativa
      indiceCalor = 0;       // Indice de calor

unsigned long tiempo,
              horaBombaParada,                     // A qué hora se paró la bomba a causa de un nivel bajo de agua
              horaUltimoMensajeError = 0,          // A qué hora se envió el último mensaje de error
              horaUltimaLectura = 0;               // A qué hora se leyeron por última vez los datos de los sensores

String bufferSerie,   // Linea leída por el puerto serie
       comando,
       infoSistema,  // Mensaje con la información del sistema que manda el arduino
       errorSistema;

bool hayErrores = false; // Estará en true, en caso de que algo vaya mal, bajo nivel de agua, no se detecta algun sensor

char *ptr = NULL;
int numErroresLeves,
    numErroresGraves;

// PARÁMETROS DE LA CONFIGURACIÓN
int  tiempoEsperaBomba;       // Tiempo que tarda el agua en volver al tanque de agua
bool envioErrores;            // Envio de errores periodicamente al servidor
int  tiempoEntreErrores;      // Cada cuánto tiempo enviar mensajes de error (5 segundos)
bool alarmaSonora;            // Emitir pitidos en caso de que haya algún error

//--------------------------------------------------------------------------
/**
  setup()
  Inicializar pines y variables
*/
void setup() {
  pinMode(12, OUTPUT);         //BUZZER
  pinMode(11, INPUT_PULLUP);   //AGUA
  pinMode(10, INPUT_PULLUP);   //BOMBA
  pinMode(10, OUTPUT);         //BOMBA
  pinMode(9,  INPUT_PULLUP);   //LEDs
  pinMode(9,  OUTPUT);         //LEDs

#ifdef DEBUG
  Serial.begin(9600);
#endif

  // Initialize the I2C bus (BH1750 library doesn't do this automatically)
  // On esp8266 you can select SCL and SDA pins using Wire.begin(D4, D3);
  Wire.begin();

  dht.begin(); // iniciar el sensor de humedad DHT11

  sensors.begin(); // Iniciar la librería para leer las temperaturas

  if(luxometreOK = lightMeter.begin()){
#ifdef DEBUG
    //Serial.println(F("BH1750 inicializado"));
#endif
  }

  comando = new char [256];

  //resoluciones de los sensores de temperatura
  sensors.setResolution(tempSensorAgua, 12);
  sensors.setResolution(tempSensorAire, 12);
  ConfiguracionPorDefecto();
}

//--------------------------------------------------------------------------
/**
  ConfiguracionPorDefecto()
  establece los palores por defecto de los parametros de la configuración,
  para que en caso de que no haya conexion con la Orange Pi
*/
void ConfiguracionPorDefecto() {
  tiempoEntreErrores = 30000;
  tiempoEsperaBomba = 60000;
  envioErrores = true;
  alarmaSonora = true;
}


//--------------------------------------------------------------------------
/**
  leerValores()
  lee todos los valores medidos por los sensores y cuenta los errores, en caso
  de haber alguno va encadenando los mensajes en una variable para pasárselos
  a la Orange Pi.
*/
bool leerValores() {
  bool ret = false;
  infoSistema = "Datos TFG: ";

  luz = medirLuz();
  if(luz>0){
    infoSistema += "luz=";
    infoSistema += (String)luz;
    infoSistema += ";";
  }
  else{
    numErroresLeves++;
    errorSistema += "luz=Error al obtener la luminosidad;";
    ret = true;
  }
  bool lecturaCorrectaTemp;
  temperaturaAgua = medirTemperatura(tempSensorAgua, lecturaCorrectaTemp);
  if(!lecturaCorrectaTemp || isnan(temperaturaAgua)){
    numErroresLeves++;
    errorSistema += "tempAgua=Error al obtener la temperatura del agua;";
    ret = true;
  }
  else {
    infoSistema += "tempAgua=";
    infoSistema += (String)temperaturaAgua;
    infoSistema += ";";
  }
  temperaturaAire = medirTemperatura(tempSensorAire, lecturaCorrectaTemp);
  if(!lecturaCorrectaTemp || isnan(temperaturaAire)){
    numErroresLeves++;
    errorSistema += "tempAire=Error al obtener la temperatura del aire;";
    ret = true;
  }
  else {
    infoSistema += "tempAire=";
    infoSistema += (String)temperaturaAire;
    infoSistema += ";";
  }
  if(medirNivelAgua()){ // Está por debajo de lo que debería estar
    horaBombaParada = millis();
    if(estadoBomba()){
      setBomba(false);
    }
    numErroresGraves++;
    errorSistema += "agua=El nivel del agua esta bajo;";
    agua = false;
    infoSistema += "agua=0;";
    ret = true;
  }
  else {
    agua = true;
    infoSistema += "agua=1;";
  }
  infoSistema += "bomba=";
  if(estadoBomba()){
    infoSistema += "ON;";
  }
  else if(estadoBomba() && !agua) {
    numErroresGraves++;
    errorSistema += "bomba=La bomba está encendida y el nivel del agua esta bajo;";
    ret = true;
  }
  else{
    infoSistema += "OFF;";
  }
  pH = medirPH();
  infoSistema += "pH=";
  infoSistema += (String)pH;
  infoSistema += ";";
  if(medirHumedad(humedad, indiceCalor)){
    infoSistema += "humedad="+(String)humedad+";indiceCalor="+(String)indiceCalor+";";
  }
  else{
    numErroresLeves++;
    errorSistema += "humedad=Error al obtener la humedad;";
    ret = true;
  }
  infoSistema += "LEDs=";
  if(getLeds()){
    infoSistema += "ON;";
  }
  else{
    infoSistema += "OFF;";
  }
  return ret;
}

//--------------------------------------------------------------------------
/**
  direccionesSensorTemperatura()
  Devuelve la dirección única del sensor de temperatura ubicado en el índice
  @param index del sensor en el bus
*/
void direccionesSensorTemperatura(int index) {
  uint8_t deviceAddress[8];
  sensors.getAddress(deviceAddress, index);
  for (int i = 0; i <8; i++){
    Serial.print(" ");
    Serial.print(deviceAddress[i]);
  }
  Serial.println();
}

//--------------------------------------------------------------------------
/**
  beep()
  Emite un pitido de la duración pasada en milisegundos
  @param duracion tiempo en milisegundos que debe estar sonando el buzzer
*/
void beep(int duracion) {
  digitalWrite(12, HIGH);
  delay(duracion);
  digitalWrite(12, LOW);
}

//--------------------------------------------------------------------------
/**
  notificarError()
  Notificar mediante el buzzer de que ha habido un error.
    Un beep corto por cada error leve
    Un beep largo por cada error grave
*/
void notificarError(){
  for (int i = 0; i < numErroresLeves;i++) {
    beep(10);
  }
  for (int i = 0; i < numErroresGraves; i++) {
     beep(300);
  }
}

//--------------------------------------------------------------------------
/**
  medirLuz()
  Intenta medir la luminosidad, si la primera vez no lo sonsigue, intenva
  volver a establecer la conexión con el sensor
  @return valor medido por el sensor de luminosidad BH1750
*/
uint16_t medirLuz(){
  int valor=-1;
  if(luxometreOK){
    valor = lightMeter.readLightLevel();
    if(valor >= 0){
      return valor;
    }
  }
   //volvemos a intentar conectar el sensor
  if(luxometreOK = lightMeter.begin()){
    return medirLuz();
  }
  else {
    numErroresLeves++;
    errorSistema += "luz=Error en la lectura de la luminosidad;";
  }
  return -1; //El valor no es correcto
}

//--------------------------------------------------------------------------
/**
  medirNivelAgua
  @return false si nivel de agua es el adecuado
*/
bool medirNivelAgua(){
  return digitalRead(11);
}

//--------------------------------------------------------------------------
/**
  medirTemperatura
  @return valor de la temperatura para el sensor con la dirección única del sensor
*/
float medirTemperatura(DeviceAddress direccionSensor, bool &correcto){
  sensors.requestTemperaturesByAddress(direccionSensor);
  correcto = sensors.isConnected(direccionSensor);
  return sensors.getTempC(direccionSensor);
}

//--------------------------------------------------------------------------
/**
  estadoBomba
  @return true si la bomba está encendida
*/
bool estadoBomba(){
  return digitalRead(10);
}

//--------------------------------------------------------------------------
/**
  setBomba
  Cambiar el estado de la bomba, ya sea encenderla o apagarla
  @param estado true para encender la bomba, falso para apagarla
*/
void setBomba(bool newEstado){
  if(!newEstado){
    digitalWrite(10, LOW);
#ifdef DEBUG
    Serial.println(F("Paro Bomba"));
#endif
  }
  else{
    digitalWrite(10, HIGH);
#ifdef DEBUG
    Serial.println(F("Enciendo Bomba"));
#endif

  }
}

//--------------------------------------------------------------------------
/**
  setLeds
  Cambiar el estado de los LEDs
  @param estado true para encenderlos, falso para apagarlos
*/
void setLeds(bool newEstado){
  if(!newEstado){
    digitalWrite(SISTEMALUZ, LOW);
#ifdef DEBUG
    Serial.println(F("Apago LEDs"));
#endif
  }
  else{
    digitalWrite(SISTEMALUZ, HIGH);
#ifdef DEBUG
    Serial.println(F("Enciendo LEDs"));
#endif
  }
}

//--------------------------------------------------------------------------
/**
  getLeds
  Devuelve el estado de los leds
  @return true encendidos, false apagados
*/
bool getLeds(){
  return digitalRead(SISTEMALUZ);
}

//--------------------------------------------------------------------------
/**
  medirPH
  Medir el pH del agua
  La idea es tomar 10 muestras de la entrada analógica, ordenarlas y descartar
  tanto la más alta como la más baja y calcular la media con las 8 muestras restantes
  convirtiendo este valor a voltaje en la variable pHVol, a continuación, utilizando
  la ecuación que hemos calculado con los valores de referencia de pH convertimos pHVol
  a pHValue y lo devolvemos
  @return phValue
*/
float medirPH() {
  int bufPH[10],
      tempPH;
  unsigned long int avgPH;

  for(int i=0;i<10;i++) {
    bufPH[i]=analogRead(PIN_PH);
    delay(10);
  }
  for(int i=0;i<9;i++) {
    for(int j=i+1;j<10;j++) {
      if(bufPH[i]>bufPH[j]) {
        tempPH=bufPH[i];
        bufPH[i]=bufPH[j];
        bufPH[j]=tempPH;
      }
     }
  }
  avgPH=0;
  for(int i=2;i<8;i++)
    avgPH+=bufPH[i];
  float pHVol=(float)avgPH*5.0/1024/6;
  float phValue = -6.30 * pHVol + 22.5;

  return phValue;
}

//--------------------------------------------------------------------------
/**
  medirHumedad
  Medimos la humedad
  @return nivel de humedad en tanto porciento (%)
*/
bool medirHumedad(float &humedad, float &indiceCalor){
  humedad = dht.readHumidity(); // Leemos la humedad relativa
  float t = dht.readTemperature(); // Leemos la temperatura en grados centígrados (por defecto) Loutilizaremos para obtener el indice de calor

  // Comprobamos si ha habido algún error en la lectura
  if (isnan(humedad)||isnan(t)) {
    return false;
  }

  // Calcular el índice de calor en grados centígrados
  indiceCalor = dht.computeHeatIndex(t, humedad, false);
  return true;
}

//--------------------------------------------------------------------------
/**
  loop()
  bucle que lee datos y espera instrucciones de la Orange Pi
*/
void loop() {
  tiempo = millis();
  numErroresLeves = 0;
  numErroresGraves = 0;
  if (((int)(tiempo - horaBombaParada) >= tiempoEsperaBomba) && !medirNivelAgua() && !estadoBomba()) {
      setBomba(true);
  }

  //Peparar array con mensaje de errores
  errorSistema = "Error TFG: ";
  infoSistema = "";

  hayErrores = leerValores();
  if( hayErrores){
    infoSistema += "sistema=ERR";
  }
  else {
    infoSistema += "sistema=OK";
  }

  //si está configurado el envio de errores, los hay
  if(envioErrores && hayErrores) {
    // y ha transcurrido el tiempo desde el último error enviado, enviamos el mensaje de error
    if (((unsigned long)(tiempo - horaUltimoMensajeError) >= tiempoEntreErrores)) {
      horaUltimoMensajeError = millis();
      //errorSistema = "El sistema tiene " + (String)numErroresLeves + " errores leves y " + numErroresGraves + " errores graves:\n" + errorSistema;
      Serial.println(errorSistema);
      // si está habilitada la alarma sonora, emitir un pitido. Un pitido corto por cada error leve y uno largo por cada error grave
      if(alarmaSonora){
        notificarError();
      }
    }
  }

  //Leer puerto serie, esperando instrucciones de la OrangePi
  while (Serial.available()) {
     char c = Serial.read();
     if (c==';') {
//        Serial.print("COMANDO::::::");
//        Serial.println(comando);
     } else {
        comando += c;
        continue;
     }
     delay(5);
  }
  if (comando.length() > 0) {
    ptr = strtok(comando.c_str(), " ");
    while(ptr != NULL) { //leemos la primera palabra. Separada del resto por un espacio.
        delay(10);
        //Serial.print("Primera Palabra !   ");
        //Serial.println(ptr);
      if((String)ptr == "Get") { //Si es una instruccion para lectura de datos...
        //Serial.println("LEEMOS:::Contestamos---->  ");
        Serial.println(infoSistema); //Mandamos toda la información
      }
      else if((String)ptr == "Config") { //Configuración del sistema...
//        Serial.println("Cambiamos configuración:::");
        ptr = strtok(NULL, " ");
        ptr = strtok(ptr,  "=/"); //leer los parametros configurables y sus nuevos valores

        while(ptr != NULL) {
//          Serial.print("Leido parametro  ");
//          Serial.print("  :");
          if ((String)ptr == "tiempoEsperaBomba") {
//              Serial.print(ptr);
              ptr = strtok(NULL, "=/");
              tiempoEsperaBomba = ((int)ptr) * 1000;
          }
          else if ((String)ptr == "tiempoEntreErrores") {
//              Serial.print(ptr);
              ptr = strtok(NULL, "=/");
              tiempoEntreErrores = ((int)ptr) * 1000;
          }
          else if ((String)ptr == "envioErrores") {
//              Serial.print(ptr);
              ptr = strtok(NULL, "=/");
              if((String) ptr = "si"){
                envioErrores = true;
              }
              else if((String) ptr = "no"){
                envioErrores = false;
              }
          }
          else if ((String)ptr == "alarmaSonora") {
//              Serial.print(ptr);
              ptr = strtok(NULL, "=/");
              if((String) ptr = "si"){
                alarmaSonora = true;
              }
              else if((String) ptr = "no"){
                alarmaSonora = false;
              }
          }
          else if  ((String)ptr == "estadoBomba") {
            ptr = strtok(NULL, "=/");
            if((String) ptr = "si" && !medirNivelAgua()){ // nos piden encender la bomba y el nivel del agua está bien
              setBomba(true);
            }
            else if((String) ptr = "no"){
              setBomba(false);
            }
          }
          else if ((String)ptr == "estadoLeds") {
            ptr = strtok(NULL, "=/");
            if((String) ptr = "si"){
              setLeds(true);
            }
            else if((String) ptr = "no"){
              setLeds(false);
            }
          }
        }
//          Serial.print(" con valor:  ");
//          Serial.println(ptr);
         ptr = strtok(NULL, "=/");
        }
//        Serial.print("CONFIGURADO!");

        ptr = strtok(NULL, " ");
     }
  }
  comando="";
}
